#!/usr/bin/env python

"""PythonicaParser: a small recursive descent parser to convert
user input into Pythonica expressions.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

I'm writing this predictive parser to fix up a bug that Christopher
Smith on Python-tutor discovered:

    http://aspn.activestate.com/ASPN/Mail/Message/726640

I must admit, though, that this isn't the prettiest code --- I'm
working with only introductory skills in recursive descent parsing.

(Hmmm.  I just read pages 1011--1018 of the Mathematica book.  A
recursive descent approach might work, but it suddenly looks very very
painful.  Hey, this is a great excuse to learn about a parser
generator like SPARK!



)

Here's the grammar that PythonicaParser tries to understand:
######################################################################

A factor is the simplest unit we can parse

    Factor -> '(' Expression ')'
    Factor -> Identifier
    Factor -> Number

'(', ')', Identifer, and Number are terminals.


Next, let's define an 'Exponential', like '2 ^ 3 ^ 4':

    Exponential -> Factor '^' Exponential
    Exponential -> Factor

Exponential groups from the right, so that 2^3^4 will parse like
2^(3^4).


Now, to handle simple math, like multiplication and division, let's
define a "Term":

    Term -> Term * Exponential
    Term -> Term / Exponential
    Term -> Term Exponential
    Term -> Exponential

And now that we have Terms, we can define Expressions that have even
lower precedence:

    Expression -> Expression + Term
    Expression -> Expression - Term
    Expression -> Term

Todo: Handle exponentiation, assignment, equality, ruledefine, ruleapply

Let's have Expression be our Start symbol.

Start -> Expression

######################################################################

We now have a grammar:

    Factor -> '(' Expression ')'
    Factor -> Identifier
    Factor -> Number

    Exponential -> Factor '^' Exponential
    Exponential -> Factor

    Term -> Term '*' Exponential
    Term -> Term '/' Exponential
    Term -> Term Exponential
    Term -> Exponential

    Expression -> Expression '+' Term
    Expression -> Expression '-' Term
    Expression -> Term

    Start -> Expression

but it's "left recursive", so I need to add some auxillary rules to
make it palatable to a recursive descent parser.


More specifcally, I'll add new rules Term1, Exponential1, and
Expression1:

    Exponential -> Factor Exponential1
    Exponential1 -> '^' Exponential
    Exponential1 -> EPSILON

    Term -> Exponential Term1
    Term1 -> '*' Exponential Term1
    Term1 -> '/' Exponential Term1
    Term1 -> Exponential Term1
    Term1 -> EPSILON

    Expression -> Term Expression1
    Expression1 -> '+' Term Expression1
    Expression1 -> '-' Term Expression1
    Expression1 -> EPSILON


######################################################################
Here's the grammar that I finally use recursive descent techniques on.

Start -> Expression EOF

Expression -> Term Expression1
Expression1 -> '+' Term Expression1
Expression1 -> '-' Term Expression1
Expression1 -> EPSILON

Term -> Exponential Term1
Term1 -> '*' Exponential Term1
Term1 -> '/' Exponential Term1
Term1 -> Exponential Term1
Term1 -> EPSILON

Exponential -> Factor Exponential1
Exponential1 -> '^' Exponential
Exponential1 -> EPSILON

Factor -> '(' Expression ')'
Factor -> Identifier
Factor -> Number
######################################################################


"""

class LexerError(Exception): pass
class ParserError(Exception): pass

"""Token types."""
IDENTIFIER = 'identifier'
NUMBER = 'number'


######################################################################
class PythonicaParser:
    def __init__(self, lexer):
        """Takes in a scanner that provides tokens via read()."""
        self.lexer = lexer

    def parseStart(self):
        expression = parseExpression()
        self.lexer.eat(PythonicaLexer.EOF)
        return expression

    ## TODO: Add implication and other rules...


    def parseExpression(self):
        FIRST = ['(', IDENTIFIER, NUMBER]
        lookahead_type, lookahead_content = self.lexer.peek()
        content = None
        if lookahead_type in FIRST:
            term = self.parseTerm()
            return ['Expression', term, self.parseExpression1()]
        else: self.errorOut()



    def parseExpression1(self):
        FIRST = ['+', '-']
        lookahead_type, lookahead_content = self.lexer.peek()
        content = None
        if lookahead_type in FIRST:
            self.lexer.eat( (lookahead_type, lookahead_content) )
            term = self.parseTerm()
            return ['Expression1', lookahead_content, term,
                    self.parseExpression1()]
        else:
            return None


    def parseTerm(self):
        FIRST = ['(', IDENTIFIER, NUMBER]
        lookahead_type, lookahead_content = self.lexer.peek()
        if lookahead_type in FIRST:
            exponential = self.parseExponential()
            return ['Term', exponential, self.parseTerm1()]
        else:
            self.errorOut()


    def parseTerm1(self):
        FIRST1 = ['*', '/']
        FIRST2 = ['(', IDENTIFIER, NUMBER]
        lookahead_type, lookahead_content = self.lexer.peek()
        if lookahead_type in FIRST1:
            self.lexer.eat( (lookahead_type, lookahead_content) )
            exponential = self.parseExponential()
            return ['Term1', lookahead_type, exponential, self.parseTerm1()]
        elif lookahead_type in FIRST2:
            exponential = self.parseExponential()
            return ['Term1', '*', exponential, self.parseTerm1()]
        else:
            return None


    def parseExponential(self):
        factor = self.parseFactor()
        return ['Exponential', factor, self.parseExponential1()]


    def parseExponential1(self):
        FIRST = ['^']
        lookahead_type, lookahead_content = self.lexer.peek()
        if lookahead_type in FIRST:
            self.lexer.eat( (lookahead_type, lookahead_content) )
            return ['Exponential1', '^', self.parseExponential()]
        else: ## EPSILON
            return None


    def parseFactor(self):
        lookahead = self.lexer.peek()
        lookahead_type, lookahead_content = lookahead
        content = None
        if lookahead_type == '(':
            self.lexer.eat(('(', '('))
            content = ['Factor', 'GROUP', self.parseExpression()]
            self.lexer.eat((')', ')'))
        elif lookahead_type == IDENTIFIER:
            content = ['Factor', 'IDENTIFIER', lookahead_content]
            self.lexer.eat(lookahead)
        elif lookahead_type == NUMBER:
            content = ['Factor', 'NUMBER', lookahead_content]
            self.lexer.eat(lookahead)
        else: self.errorOut()
        return content


    def errorOut(self):
        """Die with the very last token that we saw; perhaps this will
        give us some idea where we're running into problems."""
        raise ParserError, ("I don't know what to do with " + str(self.lexer.peek()))
######################################################################


if __name__ == '__main__':
    import PythonicaLexer
    import sys
    import userprompt
    import pprint
    p = userprompt.UserPrompt()
    while 1:
        try:
            line = p.input()
            if not line: raise SystemExit
            lexer = PythonicaLexer.makeLexer(line)
            parser = PythonicaParser(lexer)
            print pprint.pprint(parser.parseExpression())
            sys.stdout.flush()
        except ParserError, e:
            print "Parsing error: ", e
