"""PythonicaParser.py --- A small parser for Pythonica.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

History:

    This is a revised version of the PythonicaParser using the Spark
    parser generator at:

        http://www.cpsc.ucalgary.ca/~aycock/spark/

    I had tried a previous recursive-descent parser:

        http://hkn.eecs.berkeley.edu/~dyoo/python/pythonica
              /src/pythonica-parser-0.02.tar.gz

    but stopped midway because I realized that it would be a lot of
    work.  Dunno if I'll have time to finish the recursive descent
    parser, but I'll leave it there just in case.


Introduction:

    This is a parser that tries to understand a limited subset of the
    Mathematica language.  It's designed to be fairly easy to plug
    with  Pythonica symbolic math system at.

        http://www.strout.net/python/pythonica.html
    

Grammar:

    Here is the grammar that we want to parse:

        (note: the rules are in descending precedence)

    Expr                        ::= ReplaceAllExpr $

    ReplaceAllExpr              ::= ReplaceAllExpr /. RuleExpr
                                  | RuleExpr

    RuleExpr                    ::= RuleExpr -> EqualExpr
                                  | EqualExpr

    EqualExpr                   ::= EqualExpr == SetExpr
                                  | SetExpr

    SetExpr                     ::= SetExpr = UnsetExpr
                                  | UnsetExpr

    UnsetExpr                   ::= UnsetExpr =. PlusExpr
                                  | PlusExpr

    PlusExpr                    ::= PlusExpr + MinusExpr
                                  | MinusExpr

    MinusExpr                   ::= MinusExpr - TimesExpr
                                  | TimesExpr

    TimesExpr                   ::= TimesExpr * DivideExpr
                                  | TimesExpr DivideExpr
                                  | DivideExpr

    DivideExpr                  ::= DivideExpr / PowerExpr
                                  | PowerExpr

    PowerExpr                   ::= AtomExpr ^ PowerExpr
                                  | AtomExpr

    AtomExpr                    ::= identifier
                                  | number
                                  | ( Expr )
                                  | FunctionExpr

    FunctionExpr                ::= identifier [ Expr ]


Usage:

    To see how it works, just run this as a main program, and it will
    display how it is interpreting expressions.

"""

from spark import GenericParser
import PythonicaLexer

"""Our own Exception class for catching weird parse errors."""
class ParserError(Exception): pass


def parse(s):
    """Given a string, returns an AST of the parsed expression."""
    p = ExprParser()
    try:
        return p.parse(PythonicaLexer.tokenize(s))
    except Exception, e:
        raise ParserError, e


class ExprParser(GenericParser):
    """The definition of the grammar, entwined with actions that we
    perform when the parser knows enough to construct a node.  This is
    meant to be used primarily by the Spark parser generator."""

    def __init__(self, start='Expr'):
        GenericParser.__init__(self, start)

    def p_Expr(self, args):
        ' Expr ::= ReplaceAllExpr '
        return args[0]

    def p_ReplaceAllExpr_1(self, args):
        ' ReplaceAllExpr ::= ReplaceAllExpr /. RuleExpr '
        return AST(type=args[1], left=args[0], right=args[2])

    def p_ReplaceAllExpr_2(self, args):
        ' ReplaceAllExpr ::= RuleExpr '
        return args[0]

    ## Note: I feel a little hesitant here; I don't know enough of
    ## Mathematica syntax to know if this is legal or not.
    def p_RuleExpr_1(self, args):
        ' RuleExpr ::= RuleExpr -> EqualExpr '
        return AST(type=args[1], left=args[0], right=args[2])

    ## Note: I feel a little hesitant here; I don't know enough of
    ## Mathematica syntax to know if this is legal or not.
    def p_RuleExpr_2(self, args):
        ' RuleExpr ::= EqualExpr '
        return args[0]
    
    ## Note: I feel a little hesitant here; I don't know enough of
    ## Mathematica syntax to know if this is legal or not.
    def p_EqualExpr_1(self, args):
        ' EqualExpr ::= EqualExpr == SetExpr '
        return AST(type=args[1], left=args[0], right=args[2])

    ## Note: I feel a little hesitant here; I don't know enough of
    ## Mathematica syntax to know if this is legal or not.
    def p_EqualExpr_2(self, args):
        ' EqualExpr ::= SetExpr '
        return args[0]

    ## Note: I feel a little hesitant here; I don't know enough of
    ## Mathematica syntax to know if this is legal or not.
    def p_SetExpr_1(self, args):
        ' SetExpr ::= SetExpr = UnsetExpr '
        return AST(type=args[1], left=args[0], right=args[2])
        
    ## Note: I feel a little hesitant here; I don't know enough of
    ## Mathematica syntax to know if this is legal or not.
    def p_SetExpr_2(self, args):
        ' SetExpr ::= UnsetExpr '
        return args[0]
        
    ## Note: I feel a little hesitant here; I don't know enough of
    ## Mathematica syntax to know if this is legal or not.
    def p_UnsetExpr_1(self, args):
        ' UnsetExpr ::= UnsetExpr =. PlusExpr '
        return AST(type=args[1], left=args[0], right=args[2])

    ## Note: I feel a little hesitant here; I don't know enough of
    ## Mathematica syntax to know if this is legal or not.
    def p_UnsetExpr_2(self, args):
        ' UnsetExpr ::= PlusExpr '
        return args[0]

    def p_PlusExpr_1(self, args):
        ' PlusExpr ::= PlusExpr + MinusExpr '
        return AST(type=args[1], left=args[0], right=args[2])

    def p_PlusExpr_2(self, args):
        ' PlusExpr ::= MinusExpr '
        return args[0]

    def p_MinusExpr_1(self, args):
        ' MinusExpr ::= MinusExpr - TimesExpr '
        return AST(type=args[1], left=args[0], right=args[2])

    def p_MinusExpr_2(self, args):
        ' MinusExpr ::= TimesExpr '
        return args[0]

    def p_TimesExpr_1(self, args):
        ' TimesExpr ::= TimesExpr * DivideExpr '
        return AST(type=args[1], left=args[0], right=args[2])

    def p_TimesExpr_2(self, args):
        ' TimesExpr ::= TimesExpr DivideExpr '
        return AST(type=PythonicaLexer.Token(type='*'),
                   left=args[0], right=args[1])

    def p_TimesExpr_3(self, args):
        ' TimesExpr ::= DivideExpr '
        return args[0]

    def p_DivideExpr_1(self, args):
        ' DivideExpr ::= DivideExpr / PowerExpr '
        return AST(type=args[1], left=args[0], right=args[2])

    def p_DivideExpr_2(self, args):
        ' DivideExpr ::= PowerExpr '
        return args[0]
    
    def p_PowerExpr_1(self, args):
        ' PowerExpr ::= AtomExpr ^ PowerExpr '
        return AST(type=args[1], left=args[0], right=args[2])

    def p_PowerExpr_2(self, args):
        ' PowerExpr ::= AtomExpr '
        return args[0]
    
    def p_AtomExpr_1(self, args):
        ' AtomExpr ::= identifier '
        return makeASTLeaf(type='identifier', attr=args[0].attr)

    def p_AtomExpr_2(self, args):
        ' AtomExpr ::= number '
        return makeASTLeaf(type='number', attr=args[0].attr)

    def p_AtomExpr_3(self, args):
        ' AtomExpr ::= ( Expr ) '
        return args[1]

    def p_AtomExpr_4(self, args):
        ' AtomExpr ::= FunctionExpr '
        return args[0]

    def p_FunctionExpr(self, args):
        ' FunctionExpr ::= identifier [ Expr ] '
        return AST(type=PythonicaLexer.Token(type='function',
                                             attr=args[0].attr),
                   children=[args[2]])


class AST:
    def __init__(self, type, left=None, right=None, children=()):
        self.type = type.type
        self.attr = type.attr
        self._kids = list(children)
        self.left = left
        self.right = right
        if left: self._kids.append(left)
        if right: self._kids.append(right)

    def __getitem__(self, i):
        return self._kids[i]

    def isLeaf(self):
        return len(self._kids) == 0

    def __str__(self):
        if self.isLeaf(): return "%s" % self.attr
        child_list = map(str, self._kids)
        if self.attr == None:
            return "%s[%s]" % (self.type, ', '.join(child_list))
        else:
            return "%s[%s]" % (self.attr, ', '.join(child_list))


def printASTStructure(ast, level=0):
    """A quicky debug function that displays the way the AST's
    internal structure holds together."""
    print '\t' * level,
    print "%s (%s)" % (ast.type, ast.attr)
    for i in ast:
        printASTStructure(i, level+1)


def makeASTLeaf(type, attr):
    """A small utility function to make it easy to make AST leaves
    with a given type and attribute."""
    return AST(PythonicaLexer.Token(type, attr))


def test():
    """A small driver program to play with the parser."""
    import readline
    import pprint
    print "This is a test harness for PythonicaParser.  Please enter"
    print "expressions at the prompt.  To quit, please type 'quit'.\n"
    while 1:
        try:
            s = raw_input('[PythonicaParser] >>> ')
            if s == 'quit':
                raise SystemExit
            elif s:
                ast = parse(s)
                print str(s)
                print "Here's another view of the structure:"
                printASTStructure(ast)
        except ParserError, e:
            print "Parsing error:", e
        except (EOFError, SystemExit, KeyboardInterrupt):
            print "\nGoodbye!"
            raise SystemExit

if __name__ == '__main__':
    test()
